package de.metager.metagerapp;

import android.content.Context;
import android.content.Intent;
import android.content.SharedPreferences;
import android.net.Uri;
import android.util.Log;
import android.view.inputmethod.InputMethodManager;
import android.webkit.CookieManager;
import android.webkit.WebView;
import android.webkit.WebViewClient;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import de.metager.metagerapp.tokens.TokenManager;
import de.metager.metagerapp.tokens.TokenUpdater;

public class MGWebviewClient extends WebViewClient {
    private final SharedPreferences prefs;
    private final Map<String,String> supported_locales;
    private final WebView webview;
    private final InputMethodManager inputManager;
    private final Context context;
    private final TokenManager tokenManager;
    private final String baseURL;

    public MGWebviewClient(SharedPreferences prefs, Map<String, String> supported_locales, WebView webview, Context context, String baseURL) {
        this.prefs = prefs;
        this.supported_locales = supported_locales;
        this.webview = webview;
        this.context = context;
        this.inputManager = (InputMethodManager)context.getSystemService(Context.INPUT_METHOD_SERVICE);
        this.tokenManager = new TokenManager(context);
        this.baseURL = baseURL;
    }

    @Override
    public void onPageFinished(WebView view, String url) {
        // On specific URLs we have to make sure to sync Cookies so that they can be used again
        // on the next start of the app
        if(url.contains("/meta/key") || url.contains("/meta/settings")) {
            CookieManager.getInstance().flush();
        }

        if(url.contains("/keys/key/enter")){
            view.loadUrl("javascript:(function f() { document.getElementById('scan-qr-option').remove(); } )()");
        }

        try {
            tokenManager.recycleTokens(new URL(url));
        } catch (MalformedURLException e) {
            Log.e("metager", "Cannot parse URL: " + url);
        }

        // Fill Up any missing tokens
        TokenUpdater tokenUpdater = new TokenUpdater(tokenManager, baseURL);
        tokenUpdater.execute();



        Pattern locale_pattern = Pattern.compile("(https?://)([^/]*)(/.*)?/lang(\\?.*|$)");
        Matcher locale_matcher = locale_pattern.matcher(url);
        if(locale_matcher.matches()){
            String host = locale_matcher.group(2);
            String locale = locale_matcher.group(3);
            if(locale != null){
                locale = locale.replaceFirst("^/", "");
            }else{
                if(host != null && host.equals("metager.de")){
                    locale = "de-DE";
                } else if (host != null && host.equals("metager.org")) {
                    locale = "en-US";
                }
            }
            // Convert locale if shorted
            for(Map.Entry<String, String> entry : supported_locales.entrySet()){
                if(entry.getValue().equals(locale)){
                    locale = entry.getKey();
                }
            }
            // Check if current locale is the same as stored locale
            String stored_locale = prefs.getString("locale", null);

            if(stored_locale == null || !stored_locale.equals(locale)){
                prefs.edit().putString("locale", locale).apply();
            }
        }

        // Show keyboard on the startpage
        if(url.matches("^https://metager\\.[a-z]{2,3}/(\\?.*)?$")){
            inputManager.showSoftInput(webview, InputMethodManager.SHOW_IMPLICIT);
            webview.evaluateJavascript("document.querySelector('input[type=search]').focus();", null);
            webview.evaluateJavascript("document.getElementById('plugin-btn').outerHTML = '';", null);
        }
    }

    @Override
    public boolean shouldOverrideUrlLoading(WebView view, String url){
        List<String> allowed_hosts = new ArrayList<>();
        allowed_hosts.add("metager.de");
        allowed_hosts.add("metager.org");
        allowed_hosts.add("proxy.metager.de");

        String host;
        Pattern host_regex = Pattern.compile("^https?://([^/]+).*");
        Matcher host_matcher = host_regex.matcher(url);
        if(host_matcher.matches()) {
            host = host_matcher.group(1);
        }else{
            host = "";
        }

        //noinspection ConstantConditions
        if(BuildConfig.BUILD_TYPE.equals("debug")){
            allowed_hosts.add("metager3.de");
            // Add ngrok hosts to allowed list for debug apps
            if(host != null && host.matches(".*\\.ngrok-free\\.app")) {
                allowed_hosts.add(host);
            }
        }

        if(allowed_hosts.contains(host)) {
            if(host != null && !host.equals("proxy.metager.de")){
                try {
                    URL target_url = new URL(url);
                    tokenManager.validateCookies(target_url);
                    if(tokenManager.tokensNeededForRequest(target_url)){
                        Map<String,String> tokensHeader = tokenManager.getRequiredTokensAsHeader(target_url);
                        view.loadUrl(url, tokensHeader);
                    }
                } catch (MalformedURLException e) {
                    System.err.println("Cannot parse URL: " + url);
                }
            }
            return false;
        }else{
            Intent i = new Intent(Intent.ACTION_VIEW, Uri.parse(url));
            context.startActivity(i);
            return true;
        }
    }
}
