package de.metager.metagerapp;

import android.annotation.SuppressLint;
import android.app.DownloadManager;
import android.content.Context;
import android.content.SharedPreferences;
import android.content.pm.ActivityInfo;
import android.content.pm.PackageManager;
import android.net.ConnectivityManager;
import android.net.NetworkInfo;
import android.os.Bundle;
import android.telephony.TelephonyManager;
import android.view.View;
import android.webkit.CookieManager;
import android.webkit.WebSettings;
import android.webkit.WebView;
import android.widget.Toast;

import androidx.activity.OnBackPressedCallback;
import androidx.annotation.NonNull;
import androidx.appcompat.app.AppCompatActivity;

import java.util.HashMap;
import java.util.Locale;
import java.util.Map;

public class MainActivity extends AppCompatActivity {
    private WebView webview;
    private MGWebviewClient webviewClient;
    private MGWebChromeClient webChromeClient;

    @SuppressLint("SetJavaScriptEnabled")
    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.activity_main);

        this.checkForUpdates();



        webview = findViewById(R.id.webview);
        this.enableBackGestures();
        webview.getSettings().setJavaScriptEnabled(true);

        if(BuildConfig.DEBUG){
            WebView.setWebContentsDebuggingEnabled(true);
            webview.getSettings().setUserAgentString("MetaGer Development");
        }

        // Check current connectivity
        ConnectivityManager cm = (ConnectivityManager)this.getSystemService(Context.CONNECTIVITY_SERVICE);
        NetworkInfo activeNetwork = cm.getActiveNetworkInfo();
        if(null != activeNetwork && isConnectionFast(activeNetwork.getType(), activeNetwork.getSubtype())){
            webview.getSettings().setCacheMode(WebSettings.LOAD_DEFAULT);
        }else{
            webview.getSettings().setCacheMode(WebSettings.LOAD_CACHE_ELSE_NETWORK);
        }

        // Configure Cookie Management
        CookieManager.getInstance().setAcceptThirdPartyCookies(webview, false);

        webview.requestFocus(View.FOCUS_DOWN);

        // Check if user has a language preference set
        SharedPreferences prefs = this.getSharedPreferences(BuildConfig.APPLICATION_ID, Context.MODE_PRIVATE);
        String lang = prefs.getString("lang", null);
        String locale = prefs.getString("locale", null);

        // MetaGer now supports traditional locales
        // We migrate the settings
        if(lang != null){
            switch(lang){
                case "german":
                case "spanish":
                    locale = "de-DE";
                    prefs.edit().putString("locale", locale).apply();
                    break;
                case "english":
                    locale = "en-US";
                    prefs.edit().putString("locale", locale).apply();
                    break;
            }
            prefs.edit().remove("lang").apply();
        }
        Map<String, String> supported_locales = new HashMap<>();
        supported_locales.put("de-DE", "de-DE");
        supported_locales.put("en-IE", "ie");
        supported_locales.put("en-UK", "uk");
        supported_locales.put("en-US", "en-US");
        if(locale == null){
            Locale default_locale = Locale.getDefault();

            if(supported_locales.containsKey(default_locale.getLanguage() + "-" + default_locale.getCountry())){
                // Locale is supported by MetaGer: Set it as default
                locale = supported_locales.get(default_locale.getLanguage() + "-" + default_locale.getCountry());
            } else if (default_locale.getLanguage().equals("de")) {
                // We do not support the default country: Default to Germany
                locale = "de-DE";
            } else if (default_locale.getLanguage().equals("en")) {
                // We do not support the default country: Default to US
                locale = "en-US";
            }else {
                // Global default is US
                locale = "en-US";
            }
            prefs.edit().putString("locale", locale).apply();
        }
        String url = getBaseUrl() + "/" + supported_locales.get(locale);

        DownloadManager downloadManager = (DownloadManager) getSystemService(DOWNLOAD_SERVICE);
        webview.setDownloadListener(new DownloadListener(downloadManager));
        webviewClient = new MGWebviewClient(prefs, supported_locales, webview, this, url);
        webview.setWebViewClient(webviewClient);
        this.webChromeClient = new MGWebChromeClient(findViewById(R.id.progressBar), this);
        webview.setWebChromeClient(this.webChromeClient);

        if(savedInstanceState == null) {
            // Override URL Loading is not called when manually calling loadUrl
            // So we manually call it here
            webviewClient.shouldOverrideUrlLoading(webview, url);
            webview.loadUrl(url);
        }else{
            webview.restoreState(savedInstanceState);
        }
    }

    private void checkForUpdates() {

        String appID = BuildConfig.APPLICATION_ID;

        //noinspection ConstantConditions
        if(!appID.equals("de.metager.metagerapp.manual")){
            return;
        }

        //long updateCheckEvery = 2 * 60 * 60 * 1000;
        long updateCheckEvery = 60 * 1000;

        SharedPreferences prefs = this.getPreferences(Context.MODE_PRIVATE);
        long lastUpdateCheck = prefs.getLong("lastupdatecheck", 0);
        long now = System.currentTimeMillis();

        if((now - lastUpdateCheck) > updateCheckEvery) {
            UpdateChecker updateChecker = new UpdateChecker(this);
            updateChecker.execute();
            SharedPreferences.Editor editor = prefs.edit();
            editor.putLong("lastupdatecheck", System.currentTimeMillis());
            editor.apply();
        }
    }

    private String getBaseUrl() {
        try {
            ActivityInfo ai = getPackageManager().getActivityInfo(getComponentName(), PackageManager.GET_META_DATA);
            Bundle metaData = ai.metaData;
            return metaData.getString("de.metager.metagerapp.BASE_URL");
        } catch (PackageManager.NameNotFoundException e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    protected void onSaveInstanceState(@NonNull Bundle outState) {
        super.onSaveInstanceState(outState);
        webview.saveState(outState);
    }

    @Override
    protected void onRestoreInstanceState(Bundle savedInstanceState) {
        super.onRestoreInstanceState(savedInstanceState);
        webview.restoreState(savedInstanceState);
    }

    @Override
    protected void onResume() {
        super.onResume();
    }

    @Override
    protected void onPause() {
        super.onPause();
    }

    private boolean isConnectionFast(int type, int subType) {
        if(type==ConnectivityManager.TYPE_WIFI){
            return true;
        }else if(type==ConnectivityManager.TYPE_MOBILE){
            switch(subType){
                case TelephonyManager.NETWORK_TYPE_1xRTT:
                case TelephonyManager.NETWORK_TYPE_EDGE:
                    return false; // ~ 50-100 kbps
                case TelephonyManager.NETWORK_TYPE_CDMA:
                    return false; // ~ 14-64 kbps
                case TelephonyManager.NETWORK_TYPE_EVDO_0:
                    return true; // ~ 400-1000 kbps
                case TelephonyManager.NETWORK_TYPE_EVDO_A:
                    return true; // ~ 600-1400 kbps
                case TelephonyManager.NETWORK_TYPE_GPRS:
                    return false; // ~ 100 kbps
                case TelephonyManager.NETWORK_TYPE_HSDPA:
                    return true; // ~ 2-14 Mbps
                case TelephonyManager.NETWORK_TYPE_HSPA:
                    return true; // ~ 700-1700 kbps
                case TelephonyManager.NETWORK_TYPE_HSUPA:
                    return true; // ~ 1-23 Mbps
                case TelephonyManager.NETWORK_TYPE_UMTS:
                    return true; // ~ 400-7000 kbps
                /*
                 * Above API level 7, make sure to set android:targetSdkVersion
                 * to appropriate level to use these
                 */
                case TelephonyManager.NETWORK_TYPE_EHRPD: // API level 11
                    return true; // ~ 1-2 Mbps
                case TelephonyManager.NETWORK_TYPE_EVDO_B: // API level 9
                    return true; // ~ 5 Mbps
                case TelephonyManager.NETWORK_TYPE_HSPAP: // API level 13
                    return true; // ~ 10-20 Mbps
                case TelephonyManager.NETWORK_TYPE_IDEN: // API level 8
                    return false; // ~25 kbps
                case TelephonyManager.NETWORK_TYPE_LTE: // API level 11
                    return true; // ~ 10+ Mbps
                // Unknown
                case TelephonyManager.NETWORK_TYPE_UNKNOWN:
                default:
                    return false;
            }
        }else{
            return false;
        }
    }

    public void enableBackGestures() {
        OnBackPressedCallback callback = new OnBackPressedCallback(true) {
            @Override
            public void handleOnBackPressed() {
                if (webview.canGoBack()) {
                    webview.goBack();
                } else {
                    finish();
                    System.exit(0);
                }
            }
        };
        getOnBackPressedDispatcher().addCallback(this, callback);
    }

    @Override
    public void onRequestPermissionsResult(int requestCode, String[] permissions, int[] grantResults) {
        if(requestCode == MGWebChromeClient.STORAGE_PERMISSION_CODE){
            if (grantResults.length > 0 && grantResults[0] == PackageManager.PERMISSION_GRANTED) {
                this.webChromeClient.openFileExplorer();
                //Displaying a toast
//                Toast.makeText(this, "Permission granted now you can read the storage", Toast.LENGTH_LONG).show();
            } else {
                //Displaying another toast if permission is not granted
                Toast.makeText(this, R.string.storage_denied, Toast.LENGTH_LONG).show();
            }
        }
        super.onRequestPermissionsResult(requestCode, permissions, grantResults);
    }
}
