package de.metager.metagerapp.tokens;

import android.content.Context;
import android.content.SharedPreferences;
import android.util.ArrayMap;
import android.util.Log;
import android.webkit.CookieManager;

import com.google.gson.JsonArray;
import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import com.google.gson.JsonParseException;
import com.google.gson.JsonParser;

import java.net.URL;
import java.net.URLDecoder;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.TimeUnit;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import de.metager.metagerapp.BuildConfig;

/**
 * Class to manage Tokens in the App which are used to provide
 * an anonymous way to use a MetaGer Key anonymously.
 */
public class TokenManager {
    public final Context context;
    private final int cookieMaxAge = 60 * 60 * 24 * 365 * 5; // Store Cookies for five years
    private final long first_usage;
    private String key;
    private boolean keyInBrowser = false; // Whether the app has stored the key in the browser
    private int cost = 1; // Default cost for each search
    private int required_tokens = 0;    // Required Tokens for the next search

    private final ArrayList<Token> tokens;
    private int keycharge;

    public TokenManager(Context context) {
        this.context = context;

        SharedPreferences prefs = context.getSharedPreferences(BuildConfig.APPLICATION_ID, Context.MODE_PRIVATE);
        String first_use_key = "tokenmanager_first_usage";
        if (prefs.contains(first_use_key)) {
            this.first_usage = prefs.getLong("tokenmanager_first_usage", 0);
        } else {
            first_usage = System.currentTimeMillis();
            prefs.edit().putLong(first_use_key, first_usage).apply();
        }

        this.keycharge = prefs.getInt("tokenmanager_keycharge", 0);
        tokens = new ArrayList<>();
        loadKey();
        loadTokens();
    }

    /**
     * A MetaGer Key is stored by the browser as cookie and sent to
     * the server on every request.
     * When we are using tokens, the Cookie should only be sent to the
     * server for the paths where the user can view information about
     * the cookie or remove/set any key.
     * <p>
     * For all other paths the Cookie will be removed from the requests
     * and a anonymous token is added to the request for searches.
     *
     * @param url The url which is currently loaded
     */
    public void validateCookies(URL url) {
        CookieManager cookieManager = CookieManager.getInstance();
        String cookieDomain = url.getProtocol() + "://" + url.getHost();
        if (url.getPort() != -1) {
            cookieDomain += ":" + url.getPort();
        }
        Map<String, String> cookies = parseCookies(cookieDomain);

        if(key != null) {
            String tokenauthorization = "empty";
            if (keycharge > 0) {
                tokenauthorization = "low";
            }
            if (keycharge > 30) {
                tokenauthorization = "full";
            }
            cookieManager.setCookie(cookieDomain, "tokenauthorization=" + tokenauthorization + "; Path=/; Max-Age=" + cookieMaxAge + "; HttpOnly; SameSite=Strict");
            cookieManager.flush();
        }else if(cookies.containsKey("tokenauthorization") && this.tokens.size() < this.required_tokens){
            // Remove tokenauthorization cookie if present and no key is defined
            cookieManager.setCookie(cookieDomain, "tokenauthorization=deleted; Path=/; Max-Age=0; HttpOnly; SameSite=Strict");
            cookieManager.flush();
        }

        if (cookies.containsKey("cost")) {
            String cost = cookies.get("cost");
            try {
                assert cost != null;
                this.cost = Integer.parseInt(cost);
                this.required_tokens = this.cost;
            } catch (Exception e) {
            }
        }

        // Making the key removable in the webview needs some extra attention
        // since the App would just set the internally stored key over and over again
        // when it is missing in the browser
        if (this.keyInBrowser && !cookies.containsKey("key")) {
            this.removeKey();
            this.keyInBrowser = false;
        }
        // Store the Cookie if one is set for this request
        if (cookies.containsKey("key")) {
            this.keyInBrowser = true;
            storeKey(cookies.get("key"));
        }
        if (cookies.containsKey("key") && !isCookieManagementPath(url)) {
            // Remove Key Cookie for non management paths
            cookieManager.setCookie(cookieDomain, "key=; Path=/; Max-Age=0");
            cookieManager.flush();
            this.keyInBrowser = false;
        } else if (!cookies.containsKey("key") && isCookieManagementPath(url) && key != null) {
            // Set Key Cookie for management paths
            cookieManager.setCookie(cookieDomain, "key=" + key + "; Path=/; Max-Age=" + cookieMaxAge + "; HttpOnly; SameSite=Strict");
            cookieManager.flush();
            this.keyInBrowser = true;
        }
    }

    /**
     * Returns a header for the webview with the amount of currently required tokens
     *
     * @param url URL which might consume a token
     */
    public Map<String, String> getRequiredTokensAsHeader(URL url) {
        Map<String, String> tokenHeader = new ArrayMap<>();
        if (!tokensNeededForRequest(url)) {
            return tokenHeader;
        }

        // Need a case if we do not have enough Tokens in store
        if (tokens.size() < this.required_tokens) {
            // Fill Up any missing tokens
            TokenUpdater tokenUpdater = new TokenUpdater(this, url.getProtocol() + "://" + url.getHost());
            tokenUpdater.execute();
            try {
                tokenUpdater.get(30, TimeUnit.SECONDS);
            } catch (Exception e) {
                Log.e("Tokenmanager", e.getMessage());
            }
        }

        if (tokens.size() < this.required_tokens) {
            if(key != null) {
                // If there aren't enough tokens here add the key as cookie and exit. Probably an empty key
                CookieManager cookieManager = CookieManager.getInstance();
                cookieManager.setCookie(url.getProtocol() + "://" + url.getHost(), "key=" + key + "; Max-Age=" + cookieMaxAge + "; HttpOnly; SameSite=Strict");
                cookieManager.flush();
            }
            return tokenHeader;
        }

        JsonArray headersForWebview = new JsonArray();
        for (int i = 0; i < required_tokens; i++) {
            Token token = tokens.remove(0);
            headersForWebview.add(token.toJSON());
        }

        String encodedHeader = headersForWebview.toString();
        tokenHeader.put("tokens", encodedHeader);
        this.saveTokens();
        required_tokens = 0;
        CookieManager cookieManager = CookieManager.getInstance();
        cookieManager.setCookie(url.getProtocol() + "://" + url.getHost(), "cost=deleted; Max-Age=0; HttpOnly; SameSite=Strict");
        cookieManager.flush();
        return tokenHeader;
    }

    public void addToken(Token token) {
        this.tokens.add(token);
    }

    public void saveTokens() {
        // Store updated Token List in App Storage
        SharedPreferences prefs = context.getSharedPreferences(BuildConfig.APPLICATION_ID, Context.MODE_PRIVATE);

        JsonArray jsonTokens = new JsonArray();
        for (Token token : tokens) {
            jsonTokens.add(token.toJSON());
        }
        prefs.edit().putString("tokens", jsonTokens.toString()).apply();
    }

    public void loadTokens() {
        SharedPreferences prefs = context.getSharedPreferences(BuildConfig.APPLICATION_ID, Context.MODE_PRIVATE);

        String tokensString = prefs.getString("tokens", null);
        if (tokensString == null) {
            return;
        }
        JsonArray tokensJson = JsonParser.parseString(tokensString).getAsJsonArray();
        for (JsonElement token : tokensJson.asList()) {
            JsonObject tokenJson = token.getAsJsonObject();
            Token loadedToken = Token.fromJSON(tokenJson);
            if (loadedToken != null) {
                tokens.add(loadedToken);
            }
        }
    }

    public boolean isTokenPath(URL url) {
        return url.getPath().matches(".*/meta/meta.ger3$");
    }

    /**
     * Checks if new tokens are required and returns how many are needed
     *
     * @return Number of tokens that need to be generated
     */
    public int requiredTokens() {
        if (tokens.size() >= (calcMaxTokenCost() * 2)) {
            return 0;
        }
        long millisInDay = 1000L * 60 * 60 * 24;
        int numberOfDays = (int) Math.floor((System.currentTimeMillis() - first_usage) / millisInDay);
        int maxTokens = Math.min(Math.max(numberOfDays, calcMaxTokenCost() * 2), 25);
        maxTokens = Math.min(maxTokens, calcMaxTokenCost() * 5);

        return Math.min(Math.max(maxTokens - tokens.size(), 0), 10);
    }

    /**
     * Returns the maximum cost for a MetaGer search
     * with the current settings
     *
     * @return Maximum Token cost for user with current settings
     */
    public int calcMaxTokenCost() {
        return cost;
    }

    private boolean isCookieManagementPath(URL url) {
        if (url.getPath().matches(".*/meta/key$")) {
            // Deprecated MetaGer Key page; Will be removed soon
            return true;
        } else // Key Management Pages
            if (url.getPath().matches(".*/meta/settings$")) {
                // MetaGer Settings page
                return true;
            } else return url.getPath().matches(".*/keys/key/.*$");
    }

    private void storeKey(String key) {
        this.key = key;
        SharedPreferences prefs = context.getSharedPreferences(BuildConfig.APPLICATION_ID, Context.MODE_PRIVATE);
        prefs.edit().putString("key", key).apply();
    }

    private void removeKey() {
        this.key = null;
        this.keycharge = 0;
        SharedPreferences prefs = context.getSharedPreferences(BuildConfig.APPLICATION_ID, Context.MODE_PRIVATE);
        prefs.edit().remove("key").apply();
    }

    private void loadKey() {
        SharedPreferences prefs = context.getSharedPreferences(BuildConfig.APPLICATION_ID, Context.MODE_PRIVATE);
        key = prefs.getString("key", null);
    }

    public String getKey() {
        return key;
    }

    private Map<String, String> parseCookies(String cookieDomain) {
        CookieManager cookieManager = CookieManager.getInstance();

        Map<String, String> cookies = new HashMap<>();

        String cookieString = cookieManager.getCookie(cookieDomain);
        if (cookieString == null) {
            // No Cookie defined
            return cookies;
        }
        String[] cookieArray;
        if (cookieString.contains(";")) {
            // Multiple Cookies defined
            cookieArray = cookieString.split(";");
        } else {
            // Just a single cookie is defined
            cookieArray = new String[1];
            cookieArray[0] = cookieString;
        }

        Pattern cookiePattern = Pattern.compile("^([^=]+)=(.*)");

        for (String singleCookie : cookieArray) {
            singleCookie = singleCookie.trim();
            Matcher cookie_matcher = cookiePattern.matcher(singleCookie);
            if (cookie_matcher.matches()) {
                cookies.put(cookie_matcher.group(1), cookie_matcher.group(2));
            }
        }
        return cookies;
    }

    public boolean tokensNeededForRequest(URL target_url) {
        return isTokenPath(target_url) && this.required_tokens > 0;
    }

    public void setKeyCharge(int keycharge) {
        this.keycharge = keycharge;
        SharedPreferences prefs = context.getSharedPreferences(BuildConfig.APPLICATION_ID, Context.MODE_PRIVATE);
        prefs.edit().putInt("tokenmanager_keycharge", keycharge).apply();
    }

    public void recycleTokens(URL url) {
        String cookieDomain = url.getProtocol() + "://" + url.getHost();
        if (url.getPort() != -1) {
            cookieDomain += ":" + url.getPort();
        }
        Map<String, String> cookies = parseCookies(cookieDomain);
        if(cookies.containsKey("tokens")){
            try {
                String tokenJsonString = cookies.get("tokens");
                tokenJsonString = URLDecoder.decode(tokenJsonString);
                JsonArray tokens = JsonParser.parseString(tokenJsonString).getAsJsonArray();
                for(JsonElement tokenElement : tokens){
                    JsonObject token = tokenElement.getAsJsonObject();
                    Token newToken = Token.fromJSON(token);
                    this.tokens.add(0, newToken);
                }
            } catch (JsonParseException | IllegalStateException e) {
                Log.e("metager", "Cannot parse token json");
            }
            CookieManager cookieManager = CookieManager.getInstance();
            cookieManager.setCookie(cookieDomain, "tokens=deleted; Max-Age=0; HttpOnly; SameSite=Strict");
            cookieManager.flush();
        }
    }
}
