package de.metager.metagerapp.tokens;

import android.os.AsyncTask;
import android.util.Base64;

import com.google.gson.JsonArray;
import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import com.google.gson.JsonParser;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.security.KeyFactory;
import java.security.NoSuchAlgorithmException;
import java.security.interfaces.RSAPublicKey;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.X509EncodedKeySpec;
import java.util.ArrayList;
import java.util.Map;
import java.util.Set;

import javax.net.ssl.HttpsURLConnection;

public class TokenUpdater extends AsyncTask<String, Void, String> {

    private final TokenManager tokenManager;
    private final String baseApiURL;
    private int keycharge = -1;

    public TokenUpdater(TokenManager tokenManager, String baseURL){
        this.tokenManager = tokenManager;
        this.baseApiURL = baseURL + "/keys/api/json";
    }
    @Override
    protected String doInBackground(String... strings) {
        if(tokenManager.getKey() == null){
            return null;
        }
        int requiredTokens = tokenManager.requiredTokens();
        if(requiredTokens == 0){
            return null;
        }
        ArrayList<Token> newTokens = createNewTokens(requiredTokens);

        for(Token token : newTokens){
            tokenManager.addToken(token);
        }
        if(keycharge > -1){
            tokenManager.setKeyCharge(keycharge);
        }
        tokenManager.saveTokens();
        return null;
    }

    private ArrayList<Token> createNewTokens(int amount) {
        ArrayList<Token> newTokens = new ArrayList<>();
        if(amount <= 0){
            return newTokens;
        }
        Pubkey pubkey;
        try {
            pubkey = getPublicKey();
            for(int i = 0; i < amount; i++){
                newTokens.add(new Token(pubkey));
            }

            // We generated new Tokens. They all need to get signed by our server
            // To do so we send the blinded (encrypted) tokens to our server which signs them
            // We receive back the signed,blinded signatures and then can unblind the signature
            JsonObject signaturePayload = new JsonObject();
            signaturePayload.addProperty("key", tokenManager.getKey());
            signaturePayload.addProperty("date", pubkey.date);
            JsonArray blindedTokens = new JsonArray();
            for(Token token : newTokens){
                blindedTokens.add(token.getBlindedToken().toString());
            }
            signaturePayload.add("blinded_tokens", blindedTokens);

            URL signRequestURL = new URL(baseApiURL + "/token/sign");
            HttpsURLConnection request = (HttpsURLConnection) signRequestURL.openConnection();
            request.setRequestMethod("POST");
            request.setRequestProperty("Content-Type", "application/json");
            request.setDoOutput(true);
            try{
                request.connect();
                // Send Request Data
                try(OutputStream os = request.getOutputStream()) {
                    byte[] requestPayload = signaturePayload.toString().getBytes(StandardCharsets.UTF_8);
                    os.write(requestPayload);
                }

                if(request.getResponseCode() != 201){
                    return new ArrayList<>();
                }
                JsonObject jsonResponse = JsonParser.parseReader(new InputStreamReader((InputStream) request.getContent())).getAsJsonObject();
                this.keycharge = jsonResponse.get("charge").getAsInt();
                JsonObject signedTokens = jsonResponse.getAsJsonObject("signed_tokens");
                Set<Map.Entry<String, JsonElement>> signedTokensEntrySet = signedTokens.entrySet();
                for(Map.Entry<String, JsonElement> signedToken : signedTokensEntrySet){
                    String blindedToken = signedToken.getKey();
                    String blindedSignature = signedToken.getValue().getAsString();
                    for(Token token : newTokens){
                        if(token.getBlindedToken().toString().equals(blindedToken)){
                            token.addSignature(blindedSignature);
                        }
                    }
                }
            }finally{
                request.disconnect();
            }
        } catch (IOException | NoSuchAlgorithmException | InvalidKeySpecException e) {
            e.printStackTrace();
        }
        return newTokens;
    }

    private Pubkey getPublicKey() throws IOException, NoSuchAlgorithmException, InvalidKeySpecException {
        URL pubkeyURL = new URL(baseApiURL + "/token/pubkey");
        HttpsURLConnection request = (HttpsURLConnection) pubkeyURL.openConnection();
        try {
            request.connect();
            if (request.getResponseCode() == 200) {
                JsonObject jsonResponse = JsonParser.parseReader(new InputStreamReader((InputStream) request.getContent())).getAsJsonObject();
                String date = jsonResponse.get("date").getAsString();
                String pubkeyPCKS8 = jsonResponse.get("pubkey_pem").getAsString();
                pubkeyPCKS8 = pubkeyPCKS8.replace("-----BEGIN PUBLIC KEY-----", "")
                        .replace(System.lineSeparator(), "")
                        .replace("-----END PUBLIC KEY-----", "");
                byte[] encodedPubkey = Base64.decode(pubkeyPCKS8, Base64.DEFAULT);
                KeyFactory keyFactory = KeyFactory.getInstance("RSA");
                X509EncodedKeySpec keySpec = new X509EncodedKeySpec(encodedPubkey);
                return new Pubkey((RSAPublicKey) keyFactory.generatePublic(keySpec), date);
            } else {
                throw new InvalidKeySpecException("Failed Request to Keyserver");
            }
        }finally {
            request.disconnect();
        }
    }
}
